package devs.motorpackage.asc16;
/*
 *	Martin Lukac, 2003-2005 		*
 *	Intelligent Robotics Laboratory										*
 *	Portland State University 											*
 *	Office: 70-08 FAB, phone: 503-490-8478								*
 *	Mail:	P.O. Box 751												*
 *			Portland, OR 97207-0751										*
 *	Email:	lukacm@cecs.pdx.edu										*
 *	Updates and documentation: 	www.pdx.edu\~lukacm				*
 ************************************************************************/
//************************************************************************
// * 	This program is provided WITHOUT ANY WARRANTY either expressed
// *	or implied.  You may study, use, modify, and distribute it for
// *	any non-commercial purposes.  For any commercial use, prior 
// *	permission from the authors must be granted.
// ************************************************************************/
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Vector;

import javax.comm.CommPortIdentifier;
import javax.comm.NoSuchPortException;
import javax.comm.PortInUseException;
import javax.comm.SerialPort;
import javax.comm.UnsupportedCommOperationException;

import biocell.BioThread;
import monitors.CommCenter;
/**
 *Class implementing the interface to the ASC16 controller. 
 *Only basic functions are implemented becuse the controll is on higher level 
 */
public class ASC16{

	/**
	 * Comment for <code>mPortId</code>
	 */
	public static CommPortIdentifier mPortId = null;
	/**
	 * Comment for <code>mPort</code>
	 */
	public static SerialPort mPort = null;
	/**
	 * Comment for <code>os</code>
	 */
	public static OutputStream os = null;
	/**
	 * Comment for <code>is</code>
	 */
	public static InputStream is = null;
	
	/**
	 * Comment for <code>servos</code>
	 */
	public static Vector servos;

	/**
	 * @param portName
	 */
	public ASC16(String portName) {

		int baudRate = 9600;
		try {
			openPort(portName);
			initPort(baudRate);
			ASC16.os = getOutputStream();
			new InputReader();
		} catch (Exception e) {
			CommCenter.showError("Problem with initiation of the port ... aborting!");//e.printStackTrace();
		}
		
		//send this command to initiate the board...
		sendInt(121);
		sendInt(255);
		//tunr all servos on
		sendInt(245);
		//set the trigger level to repport
		setTriggerLevel(triggerLevel);
		servos = new Vector(1);
		busy = false;
		CommCenter.showComm("The ASC16 is up and running...");

	}

	/**
	 * Register a Servo with ASC16 controller to obtain direct feedback
	 * @param s - Servo
	 */
	public void register(Servo s){
		servos.add(s);
	}
	
	/**
	 * Sends a command to the ASC16 micro-controller 
	 * @param i - integer
	 */
	public synchronized void sendInt(int i) {
		try {
			while (busy){ wait(1);}
			busy = true;
			try{
				ASC16.os.write(i);
				CommCenter.showDebug("Sent: "+i);
			}catch (Exception e){} 
		} catch (Exception e) {
			CommCenter.showError(e);
		}
		busy = false;
	}

	/**
	 * Sends a array of integer representing a single command to the ASC16
	 * This limits the problem of command mixing - easy
	 * @param i
	 */
	public synchronized void sendInt(int [] i) {
		try {
			while (busy){ wait(1);}
			busy = true;
			for (int a = 0; a< i.length; a++)
				try{
				ASC16.os.write(i[a]);
				CommCenter.showDebug("Sent: "+i[a]);
				}catch (Exception e){} 
		} catch (Exception e) {
			CommCenter.showError(e);
		}
		busy = false;
	}

	/**
	 * @param portName
	 * @throws NoSuchPortException
	 * @throws PortInUseException
	 */
	public void openPort(String portName)
		throws NoSuchPortException, PortInUseException {
		CommCenter.showComm("Opening port..." + portName);
		try{
			ASC16.mPortId = CommPortIdentifier.getPortIdentifier(portName);
			ASC16.mPort = (SerialPort) ASC16.mPortId.open(portName, 30000);
		} catch (NoSuchPortException e){
			CommCenter.showError("Port "+portName+" does not exists");
		}
	}

	/**
	 *  Close the currently open serial port
	 */
	public void closePort() {
		CommCenter.showComm("Closing port...");
		ASC16.mPort.close();
	}

	/**
	 * @param baud
	 * @throws UnsupportedCommOperationException
	 */
	public void initPort(int baud) throws UnsupportedCommOperationException {
		CommCenter.showComm("Initializing Port...");

		// Set the baud rate and other parameters.
		ASC16.mPort.setSerialPortParams(
			baud,
			SerialPort.DATABITS_8,
			SerialPort.STOPBITS_1,
			SerialPort.PARITY_NONE);

		// Setup the flow control
		ASC16.mPort.setFlowControlMode(
			javax.comm.SerialPort.FLOWCONTROL_NONE);
	}

	/**
	 * @return OutputStream to use for communicating with the servo controller.
	 * @throws IOException
	 */
	OutputStream getOutputStream() throws IOException {
		return ASC16.mPort.getOutputStream();
	}

	/**
	 * @return InuputStream to use for communicating with the servo controller.
	 * @throws IOException
	 */
	InputStream getInputStream() throws IOException {
		return ASC16.mPort.getInputStream();
	}

	/**
	 * Sets the trigger level of the ASC16 controller 
	 * @param triggerLevel - integer [0-3]
	 */
	public void setTriggerLevel(int triggerLevel){
		this.triggerLevel = triggerLevel;
		sendInt(119);
		sendInt(triggerLevel);
	}
	
	/**
	 * 
	 * @author Lukac
	 *
	 * A threaded listener of the serial port. It seems the ASC16
	 * is almost random how it sends the ybtes back, thus this
	 * class is catching them all
	 */
	class InputReader extends BioThread{
		public InputReader(){
			super("asc16");
			try {
				ASC16.is = getInputStream();
			} catch (IOException e) {
				closePort();
				CommCenter.showError(e.getMessage());
			}
			start();
		}
		
		public void run(){
			int answer;
			while(true){
				try{
					if ((answer = is.read()) >= -1000000) //to be sure not to send noise
						CommCenter.showDebug("Got feedback form controller: "+answer);
						if (answer > 0) //suporeted are only triggers on motion completion
							for (int a = 0;a<servos.size(); a++)
								if ((((Servo)servos.elementAt(a)).getMotorNumber() + 180) == answer){
									CommCenter.showDebug("The servo "+answer+ " send a feedback");
									((Servo)servos.elementAt(a)).stateChanged();
									//answer = -100;
								}
							answer = -100;
				}catch (IOException e){
					e.printStackTrace();
				}
			}
		}
	}
	
	public static void main(String [] args){
		ASC16 asc = new ASC16("/dev/ttyS0");
		asc.sendInt(81);
		asc.sendInt(81);
	}
	
	//default trigger level is set to "Repport only"
	protected int triggerLevel = 1;
	//the blocking variable
	protected boolean busy = false;
}
