/**
 *  Lego/Java Robotics Development Kit
 *
 * @author S.Ritter  01/8/2001
 * @version 1.0
 *
 *  ALL EXAMPLES OF CODE AND/OR COMMAND-LINE INSTRUCTIONS ARE BEING
 *  PROVIDED BY SUN AS A COURTESY, "AS IS," AND SUN DISCLAIMS ANY AND
 *  ALL WARRANTIES PERTAINING THERETO, INCLUDING ANY WARRANTIES OF
 *  MERCHANTABILTY, FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.
 *  SUN IS NOT LICENSING THIS EXAMPLE FOR ANY USE OTHER THAN FOR THE
 *  EDUCATIONAL PURPOSE OF SHOWING THE FUNCTIONALITY CONTAINED
 *  THEREIN, BY WAY OF EXAMPLE.
 **/
package devs.visionpackage.org.cmotch.robovision;

import java.awt.Image;
import java.awt.image.ColorModel;
import java.awt.image.ImageObserver;
import java.awt.image.PixelGrabber;

/**
 *  Class that represents the image that was captured from the camera
 *  This includes methods to allow the caller to find out what objects
 *  were seen, if any, and what colour they were.
 **/
public class VisionImage {
  /*  Color codes for search.  Using these values we can greatly simplify
   *  the analysis process
   */
  public final static byte RED = 0;
  public final static byte GREEN = 1;
  public final static byte BLUE = 2;
  public final static byte NUMBER_OF_COLOURS = 3;
  public final static byte NO_COLOUR = 4;

  /*  Codes for the processed sub-sections of the image  */
  public final static byte EMPTY = 0;
  public final static byte EDGE = 1;
  public final static byte OCCUPIED = 2;

  /**  Parameters for color detection:
   *   - minimumPixels is the minimum number of coloured pixels we have
   *     to find before we say we have an object in view
   *   - thresholds represents the values that are used to evaluate a pixel
   *     to see if it is of the colour we want.  The lighting and the shape
   *     of the balls means that the intensity of colour varies greatly over
   *     the surface of the object.  This means that we don't get nice, pure
   *     colours, e.g blue will show up as blue, but as the brightness
   *     increases so too does the level of green.  We therefore have an
   *     array that defines the levels to work against.
   *   - The absolute array tells the system whether the values defined
   *     in the threshold array should be applied as absolute values (true)
   *     or relative values (false)
   **/
  private int minimumPixels = 25;
  private int[][] thresholds = {{130, 80, 100},
                                {60, 100, 60},
                                {60, 60, 120}};
  private boolean[][] thresholdAbsolute = {{true, true, false},
                                           {true, true, false},
                                           {true, false, true}};

  private Image image;
  private VisionObject[] visionObjects = new VisionObject[3];

  private int imageWidth;
  private int imageHeight;
  private int[] pixels;

  private boolean[] rgbSeen = new boolean[3];

  private boolean imageScanned = false;
  private int debugLevel;

  /**
   *  Constructor
   *
   * @param image The source image from the camera
   * @throws VisionException If the image is unusable
   **/
  public VisionImage(Image image) throws VisionException {
    this.image = image;
    imageWidth = image.getWidth(null);
    imageHeight = image.getHeight(null);

    if (imageWidth <= 0 || imageHeight <= 0)
      throw new VisionException("One of image dimensions <= 0");

    /*  Turn the Image object into an array of pixels which are easier to
     *  process
     */
    pixels = new int[imageWidth * imageHeight];
    PixelGrabber pg = null;

    try {
      pg = new PixelGrabber(image, 0, 0, imageWidth,
        imageHeight, pixels, 0, imageWidth);
      if (pg.grabPixels() == false)
        throw new VisionException("PG: 0 grabPixels failed " + pg.getStatus());
    } catch (Exception e) {
      throw new VisionException("PG: 1 " + e.getMessage());
    }

    /*  Make sure it worked  */
    if ((pg.status() & ImageObserver.ABORT) != 0)
      throw new VisionException("PG: 2 Error grabbing pixels");
  }

  /**
   *  Set debug level
   **/
  public void setDebugLevel(int level) {
    debugLevel = level;
  }

  /**
   *  Get the width of the image
   *
   * @return The width of the image
   **/
  public int getWidth() {
    return imageWidth;
  }

  /**
   *  Get the height of the image
   *
   * @return The height of the image
   **/
  public int getHeight() {
    return imageHeight;
  }

  /**
   *  Get the raw Image object (can be used for display)
   *
   * @return The initial image
   **/
  public Image getImage() {
    return image;
  }

  /**
   *  Get the pixel map for this image
   *
   * @return The pixel array for this image
   **/
  public int[] getPixelMap() {
    return pixels;
  }

  /**
   *  Get the value of a particular pixel
   *
   * @param x The X co-ordinate of the pixel
   * @param y The Y co-ordinate of the pixel
   **/
  public int getPixel(int x, int y) {
    return pixels[(y * imageWidth) + x];
  }

  /**
   *  Get the colours, if any, that were seen in the image.
   *  <p><b>NOTE:</b>  These vision algorithms are quite simple, so we restrict
   *  our processing to the primary colours red, blue and green.  If you put
   *  a non-primary coloured object in the field of view it may well appear
   *  as two objects in the same place.
   *
   * @returns A boolean array to indicate which of the RGB colours were seen
   **/
  public boolean[] getColoursSeen() {
    if (imageScanned)
      return rgbSeen;

    int numOfPixels = pixels.length;
    int redResult = 0;
    int blueResult = 0;
    int greenResult = 0;

    for (int i = 0; i < numOfPixels; i++) {
      if (testPixelColour(pixels[i], RED))
        redResult++;

      if (testPixelColour(pixels[i], GREEN))
        greenResult++;

      if (testPixelColour(pixels[i], BLUE))
        blueResult++;
    }

    if (redResult > minimumPixels)
      rgbSeen[RED] = true;

    if (blueResult > minimumPixels)
      rgbSeen[BLUE] = true;

    if (greenResult > minimumPixels)
      rgbSeen[GREEN] = true;

    if (debugLevel > 1)
      System.out.println("RED [" + redResult + "], BLUE [" + blueResult +
        "], GREEN [" + greenResult + "]");

    imageScanned = true;
    return rgbSeen;
  }

  /**
   *  Determine whether a pixel is of a particular colour using the
   *  thresholds array
   *
   * @param pixel The pixel to check
   * @param colour The colour to test the pixel for
   * @return Whether the pixel is of this colour or not
   **/
  public boolean testPixelColour(int pixel, byte colour) {
    /*  Separate the pixel into its RGB colour components  */
    int[] colourValues = new int[3];
    ColorModel cm = ColorModel.getRGBdefault();
    colourValues[RED] = cm.getRed(pixel);
    colourValues[GREEN] = cm.getGreen(pixel);
    colourValues[BLUE] = cm.getBlue(pixel);

    /*  First, does the colour of this pixel exceed our base threshold?  */
    if (colourValues[colour] > thresholds[colour][colour]) {
      /*  Figure out what the other two colours are  */
      byte nonColour1 = colour == GREEN ? BLUE : GREEN;
      byte nonColour2;

      if (colour + nonColour1 == NUMBER_OF_COLOURS)
        nonColour2 = RED;
      else
        nonColour2 = BLUE;

      /*  Now test the pixel against the other two colour thresholds  */
      if (testOtherColour(colourValues, colour, nonColour1))
        if (testOtherColour(colourValues, colour, nonColour2))
          return true;
    }

    return false;
  }

  /**
   *  Test one of the other colours in a pixel to see whether it meets our
   *  criteria
   *
   * @param colourValue The colour values of the pixel
   * @param colour The colour being searched for
   * @param nonColour The colour being tested
   * @return If the pixel meets our requirements
   **/
  private boolean testOtherColour(int[] colourValue, byte colour,
      byte nonColour) {
    if (thresholdAbsolute[colour][nonColour]) {
      if (colourValue[nonColour] < thresholds[colour][nonColour])
        return true;
    } else {
      if (colourValue[colour] - colourValue[nonColour] >
          thresholds[colour][nonColour])
        return true;
    }

    return false;
  }

  /**
   *  Look for coloured object.  This will be useful if there could be more
   *  than one object in the field of view.
   *
   * @param colour The colour of the object to look for
   * @returns True if an object of this colour is visible, otherwise false
   **/
  public boolean canYouSee(byte colour) {
    /*  If the image hasn't been scanned, scan it  */
    if (!imageScanned)
      getColoursSeen();
    return rgbSeen[colour];
  }

  /**
   *  Get the information about an object viewed by the robot
   *
   * @param colour The colour to look for (NO_COLOUR means any colour,
   *   which will be the first of red, blue or green that was seen
   * @returns The VisionImage object (null if no object)
   **/
  public VisionObject getVisionObject(byte colour) {
    /*  Check to see if the image has been scanned for a coloured object  */
    if (imageScanned == false)
      getColoursSeen();

    /*  If the user specified no colour set the colour to the first colour
     *  that matches positive.  If no colours seen return null.
     */
    if (colour == NO_COLOUR) {
      if (rgbSeen[RED])
        colour = RED;
      else if (rgbSeen[BLUE])
        colour = BLUE;
      else if (rgbSeen[GREEN])
        colour = GREEN;
      else
        return null;
    } else {
      if (!rgbSeen[colour])
        return null;
    }

    /*  Check to see if this colour has already been processed  */
    if (visionObjects[colour] == null)
      visionObjects[colour] = new VisionObject(this, colour);

    return visionObjects[colour];
  }
}
