/**
 *  Lego/Java Robotics Development Kit
 *
 * @author S.Ritter  15/8/2001
 * @version 1.0
 *
 *  ALL EXAMPLES OF CODE AND/OR COMMAND-LINE INSTRUCTIONS ARE BEING
 *  PROVIDED BY SUN AS A COURTESY, "AS IS," AND SUN DISCLAIMS ANY AND
 *  ALL WARRANTIES PERTAINING THERETO, INCLUDING ANY WARRANTIES OF
 *  MERCHANTABILTY, FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.
 *  SUN IS NOT LICENSING THIS EXAMPLE FOR ANY USE OTHER THAN FOR THE
 *  EDUCATIONAL PURPOSE OF SHOWING THE FUNCTIONALITY CONTAINED
 *  THEREIN, BY WAY OF EXAMPLE.
 **/
package devs.visionpackage.org.cmotch.robovision;

import java.awt.*;
import java.awt.image.*;

/**
 *  Class that represents an object seen by the robot
 **/
public class VisionObject {
  public final static int NO_DATA = -1;
  public final static int CENTRE = 0;
  public final static int LEFT = 1;
  public final static int RIGHT = 2;

  private final static int LEFT_MARGIN_DEFAULT = 33;
  private final static int RIGHT_MARGIN_DEFAULT = 66;

  private final static int[][] nonColour = {{1, 2}, {0, 2}, {1, 3}};

  private byte colour;
  private int whiteThreshold = 280;
  private int upperThreshold;
  private int lowerThreshold;

  private Point centre;
  private int left;
  private int right;
  private int top;
  private int bottom;

  private int imageWidth;
  private int leftMargin;
  private int rightMargin;
  private VisionImage visionImage;

  private boolean noData;
  private int debugLevel;

  /**
   *  Constructor
   *
   * @param visionImage The image to analyse
   * @param colour The colour of the object we're interested in
   **/
  public VisionObject(VisionImage visionImage, byte colour) {
    this.colour = colour;
    centre = new Point();
    boolean[] coloursSeen = visionImage.getColoursSeen();

    if (!coloursSeen[colour]) {
      noData = true;
      return;
    }

    imageWidth = visionImage.getWidth();
    int imageHeight = visionImage.getHeight();
    int[] pixels = visionImage.getPixelMap();

    leftMargin = (LEFT_MARGIN_DEFAULT * imageWidth) / 100;
    rightMargin = (RIGHT_MARGIN_DEFAULT * imageWidth) / 100;

    /*  Note on the co-ordinate system used.  The pixel array generated by the
     *  PixelGrabber class has the origin at the top left corner.  This means
     *  that whilst the left value is less than the right value the top value
     *  is actually less than the bottom.
     */
    left = imageWidth;
    right = 0;
    top = imageHeight;
    bottom = 0;
    ColorModel cm = ColorModel.getRGBdefault();

    /*  Loop through each pixel, row by row.  Figure out where the bounding
     *  corners of the object are
     */
    for (int y = 0; y < imageHeight; y++) {
      for (int x = 0; x < imageWidth; x++) {
        int p = (y * imageWidth) + x;

        /*  If this pixel is in our object we need to see if it is outside
         *  the currently recorded bounds.
         */
        if (visionImage.testPixelColour(pixels[p], colour)) {
          if (x < left)
            left = x;
          if (y < top)
            top = y;
          if (x > right)
            right = x;
          if (y > bottom)
            bottom = y;
        }
      }
    }

    /*  Now that we have the area enclosing the object we can figure out
     *  where the centre is
     */
    centre.x = left + ((right - left) / 2);
    centre.y = top + ((bottom - top) / 2);
  }

  /**
   *  Get the colour of this object
   *
   * @return The colour of the object
   **/
  public byte getColour() {
    return colour;
  }

  /**
   *  Get the position of the object, i.e. is it in the centre of the field
   *  of view, to the left or the right
   *
   * @returns The position of the object in the field of view
   **/
  public int getPosition() {
    if (noData)
      return NO_DATA;
    if (centre.x < leftMargin)
      return LEFT;
    if (centre.x > rightMargin)
      return RIGHT;
    return CENTRE;
  }

  /**
   *  Get the position of the left most pixel in our object
   *
   * @returns The left most pixel position for the object
   **/
  public int getLeft() {
    if (noData)
      return NO_DATA;
    return left;
  }

  /**
   *  Get the position of the right most pixel in our object
   *
   * @returns The right most pixel position for the object
   **/
  public int getRight() {
    if (noData)
      return NO_DATA;
    return right;
  }

  /**
   *  Get the position of the top most pixel in our object
   *
   * @returns The top most pixel position for the object
   **/
  public int getTop() {
    if (noData)
      return NO_DATA;
    return top;
  }

  /**
   *  Get the position of the bottom most pixel in our object
   *
   * @returns The bottom most pixel position for the object
   **/
  public int getBottom() {
    if (noData)
      return NO_DATA;
    return bottom;
  }

  /**
   *  Get the position of the centre of our object
   *
   * @returns A point which is the centre of the object, null if no data
   **/
  public Point getCentre() {
    if (noData)
      return null;
    return centre;
  }

  /**
   *  Set the left hand margin value.  This is the value that will determine
   *  whether the class decides that the object is to the left of the view
   *  or not.  Values should be expressed as a percentage of the total width
   *  of the image.  Values over 50% will be ignored.
   *
   * @param margin The percentage of the image width that is the left margin
   **/
  public void setLeftMargin(int margin) {
    /*  Silently ignore bad values  */
    if (margin < 0 || margin > 49)
      return;

    leftMargin = (visionImage.getWidth() * margin) / 100;
  }

  /**
   *  Set the right hand margin value.  This is the value that will determine
   *  whether the class decides that the object is to the right of the view
   *  or not.  Values should be expressed as a percentage of the total width
   *  of the image.  Values less than 50% or over 100% will be ignored.
   *
   * @param margin The percentage of the image width that is the left margin
   **/
  public void setRightMargin(int margin) {
    /*  Silently ignore bad values  */
    if (margin < 50 || margin > 100)
      return;

    rightMargin = (visionImage.getWidth() * margin) / 100;
  }

  /**
   *  Set the debug level to ouput information
   *
   * @param debugLevel The debug level to use 1=less messages, 3 is max
   **/
  public void setDebugLevel(int debugLevel) {
    this.debugLevel = debugLevel;
  }
}
