package mainframe;
import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.io.*;
import java.util.*;
import javax.swing.text.*;
import javax.swing.*;

/**
 * @author Martin Lukac
 * @version October 2003
 */
class Editor extends MIFrame {

	/**
	 * default constructor
	 */
	Editor() {
		super(600, 400, "Editor", true, true);

		try {
			UIManager.setLookAndFeel(
				UIManager.getCrossPlatformLookAndFeelClassName());
		} catch (Exception exc) {
			System.err.println("Error loading L&F: " + exc);
		}

		getContentPane().setLayout(new BorderLayout());

		// create the embedded JTextComponent
		editor = createEditor();

		// install the command table
		commands = new Hashtable();
		Action[] actions = getActions();
		for (int i = 0; i < actions.length; i++) {
			Action a = actions[i];
			commands.put(a.getValue(Action.NAME), a);
		}

		JScrollPane scroller = new JScrollPane();
		JViewport port = scroller.getViewport();
		port.add(editor);

		JPanel panel = new JPanel();
		panel.setLayout(new BorderLayout());
		panel.add("North", createToolbar());
		panel.add("Center", scroller);
		getContentPane().add("Center", panel);
		getContentPane().add("South", createStatusbar());
	}

	/**
	 * Fetch the list of actions supported by this editor. It is implemented to
	 * return the list of actions supported by the embedded JTextComponent
	 * augmented with the actions defined locally.
	 */
	public Action[] getActions() {
		return TextAction.augmentList(editor.getActions(), defaultActions);
	}

	/**
	 * Create an editor to represent the given document.
	 */
	protected JTextComponent createEditor() {
		JTextComponent c = new JTextArea();
		c.setDragEnabled(true);
		c.setFont(new Font("monospaced", Font.PLAIN, 12));
		return c;
	}

	/**
	 * Fetch the editor contained in this panel
	 */
	protected JTextComponent getEditor() {
		return editor;
	}


	/**
	 * Find the hosting frame, for the file-chooser dialog.
	 */
	protected Frame getFrame() {
		for (Container p = getParent(); p != null; p = p.getParent()) {
			if (p instanceof Frame) {
				return (Frame) p;
			}
		}
		return null;
	}

	/**
	 * This is the hook through which all menu items are created. It registers
	 * the result with the menuitem hashtable so that it can be fetched with
	 * getMenuItem().
	 * 
	 * @see #getMenuItem
	 */
	protected JMenuItem createMenuItem(String cmd) {
		JMenuItem mi = new JMenuItem(cmd);
		mi.setHorizontalTextPosition(JButton.RIGHT);
		String astr = cmd;
		if (astr == null) {
			astr = cmd;
		}
		mi.setActionCommand(astr);
		Action a = getAction(astr);
		if (a != null) {
			mi.addActionListener(a);
			a.addPropertyChangeListener(createActionChangeListener(mi));
			mi.setEnabled(a.isEnabled());
		} else {
			mi.setEnabled(false);
		}
		menuItems.put(cmd, mi);
		return mi;
	}

	/**
	 * Fetch the menu item that was created for the given command.
	 * 
	 * @param cmd
	 *                  Name of the action. @returns item created for the given
	 *                  command or null if one wasn't created.
	 */
	protected JMenuItem getMenuItem(String cmd) {
		return (JMenuItem) menuItems.get(cmd);
	}

	protected Action getAction(String cmd) {
		return (Action) commands.get(cmd);
	}

	protected Container getToolbar() {
		return toolbar;
	}

	protected JMenuBar getMenubar() {
		return menubar;
	}

	/**
	 * Create a status bar
	 */
	protected Component createStatusbar() {
		// need to do something reasonable here
		status = new StatusBar();
		return status;
	}

	/**
	 * Create the toolbar. By default this reads the resource file for the
	 * definition of the toolbar.
	 */
	private Component createToolbar() {
		toolbar = new JToolBar();
		toolbar.add(createTool(newAction));
		toolbar.add(createTool(openAction));
		toolbar.add(createTool(saveAction));
		toolbar.add(Box.createHorizontalGlue());
		return toolbar;
	}

	/**
	 * Hook through which every toolbar item is created.
	 */
	protected Component createTool(String key) {
		return createToolbarButton(key);
	}

	/**
	 * Create a button to go inside of the toolbar.
	 */
	protected JButton createToolbarButton(String key) {
		JButton b = new JButton(key) {
			public float getAlignmentY() {
				return 0.5f;
			}
		};
		b.setRequestFocusEnabled(false);
		b.setMargin(new Insets(1, 1, 1, 1));

		String astr = key;
		if (astr == null) {
			astr = key;
		}
		Action a = getAction(astr);
		if (a != null) {
			b.setActionCommand(astr);
			b.addActionListener(a);
		} else {
			b.setEnabled(false);
		}

		return b;
	}

	/**
	 * Take the given string and chop it up into a series of strings on
	 * whitespace boundries. This is useful for trying to get an array of
	 * strings out of the resource file.
	 */
	protected String[] tokenize(String input) {
		Vector v = new Vector();
		StringTokenizer t = new StringTokenizer(input);
		String cmd[];

		while (t.hasMoreTokens())
			v.addElement(t.nextToken());
		cmd = new String[v.size()];
		for (int i = 0; i < cmd.length; i++)
			cmd[i] = (String) v.elementAt(i);

		return cmd;
	}

	// Yarked from JMenu, ideally this would be public.
	protected PropertyChangeListener createActionChangeListener(JMenuItem b) {
		return new ActionChangedListener(b);
	}

	// Yarked from JMenu, ideally this would be public.
	private class ActionChangedListener implements PropertyChangeListener {
		JMenuItem menuItem;

		ActionChangedListener(JMenuItem mi) {
			super();
			this.menuItem = mi;
		}
		public void propertyChange(PropertyChangeEvent e) {
			String propertyName = e.getPropertyName();
			if (e.getPropertyName().equals(Action.NAME)) {
				String text = (String) e.getNewValue();
				menuItem.setText(text);
			} else if (propertyName.equals("enabled")) {
				Boolean enabledState = (Boolean) e.getNewValue();
				menuItem.setEnabled(enabledState.booleanValue());
			}
		}
	}

	private JTextComponent editor;
	private Hashtable commands;
	private Hashtable menuItems;
	private JMenuBar menubar;
	private JToolBar toolbar;
	private JComponent status;
	protected FileDialog fileDialog;

	public static final String openAction = "Open";
	public static final String newAction = "New";
	public static final String saveAction = "Save";
	public static final String exitAction = "Exit";

	/**
	 * FIXME - I'm not very useful yet
	 */
	class StatusBar extends JComponent {

		public StatusBar() {
			super();
			setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
		}

		public void paint(Graphics g) {
			super.paint(g);
		}

	}

	/**
	 * Actions defined by the Editor class
	 */
	private Action[] defaultActions =
		{
			new NewAction(),
			new OpenAction(),
			new SaveAction(),
			};

	class OpenAction extends NewAction {

		OpenAction() {
			super(openAction);
		}

		public void actionPerformed(ActionEvent e) {
			Frame frame = getFrame();
			if (fileDialog == null) {
				fileDialog = new FileDialog(frame);
			}
			fileDialog.setMode(FileDialog.LOAD);
			fileDialog.setVisible(true);

			String file = fileDialog.getFile();
			if (file == null) {
				return;
			}
			String directory = fileDialog.getDirectory();
			File f = new File(directory, file);
			if (f.exists()) {
				Document oldDoc = getEditor().getDocument();
				getEditor().setDocument(new PlainDocument());
				frame.setTitle(file);
				Thread loader = new FileLoader(f, editor.getDocument());
				loader.start();
			}
		}
	}

	class SaveAction extends NewAction {

		SaveAction() {
			super(saveAction);
		}

		public void actionPerformed(ActionEvent e) {
			Frame frame = getFrame();
			if (fileDialog == null) {
				fileDialog = new FileDialog(frame);
			}
			fileDialog.setMode(FileDialog.SAVE);
			fileDialog.setVisible(true);

			String file = fileDialog.getFile();
			if (file == null) {
				return;
			}
			String directory = fileDialog.getDirectory();
			File f = new File(directory, file);
			Document Doc = getEditor().getDocument();
			try
			{				
				FileWriter fW = new FileWriter(f);
				if (f.exists()) {
					String message = "are you sure you want "+'\n'+"to overwirte this file??";
					// Modal dialog with yes/no button
					//TODO
					//TODO: messages should be of the Jdialogpanel style
					int answer = JOptionPane.showConfirmDialog(frame, message);
					if (answer == JOptionPane.YES_OPTION) {
						try{
							fW.write(Doc.getText(Doc.getStartPosition().getOffset(), Doc.getEndPosition().getOffset()));
						} catch (BadLocationException ble){
							
						}
					}
				} else 
				try{
					fW.write(Doc.getText(Doc.getStartPosition().getOffset(), Doc.getEndPosition().getOffset()));
				} catch (BadLocationException ble){
					
				}
					fW.flush();
					fW.close();
				
			}catch (IOException ioe)
			{
				
			}
		}
	}

	class NewAction extends AbstractAction {

		NewAction() {
			super(newAction);
		}

		NewAction(String nm) {
			super(nm);
		}

		public void actionPerformed(ActionEvent e) {
			Document oldDoc = getEditor().getDocument();
			getEditor().setDocument(new PlainDocument());
			revalidate();
		}
	}

	class LoadAction extends AbstractAction {

		LoadAction() {
			super(newAction);
		}

		LoadAction(String nm) {
			super(nm);
		}

		public void actionPerformed(ActionEvent e) {
			Document oldDoc = getEditor().getDocument();
			getEditor().setDocument(new PlainDocument());
			revalidate();
		}
	}

	/**
	 * Thread to load a file into the text storage model
	 */
	class FileLoader extends Thread {

		FileLoader(File f, Document doc) {
			setPriority(4);
			this.f = f;
			this.doc = doc;
		}

		public void run() {
			try {
				
				
				status.removeAll();
				JProgressBar progress = new JProgressBar();
				progress.setMinimum(0);
				progress.setMaximum((int) f.length());
				status.add(progress);
				status.revalidate();

				Reader in = new FileReader(f);
				char[] buff = new char[4096];
				int nch;
				while ((nch = in.read(buff, 0, buff.length)) != -1) {
					doc.insertString(
						doc.getLength(),
						new String(buff, 0, nch),
						null);
					progress.setValue(progress.getValue() + nch);
				}

				status.removeAll();
				status.revalidate();

			} catch (IOException e) {
				System.err.println(e.toString());
			} catch (BadLocationException e) {
				System.err.println(e.getMessage());
			}
		}

		Document doc;
		File f;
	}

}
