/*
 * Created on Aug 16, 2005
 *
 */
package mns;

/**
 * @author martin lukac
 *
 * Interface with all basic definitions of the Command like objects. 
 */
interface Commandable {
	/**
	 * This execution is the simplest. We write the command and wait for
	 * feedback <code>SINGLE_ACT</code>
	 */
	final public static int SINGLE_ACT = 0;

	/**
	 * A periodic activity such as Speaking or Walking that requires a temporal
	 * synchronization of mulltiple BioDevices. This is done via the usage of
	 * the BioDeviceListener interface allowing to listen to each BioDevice
	 * Individualy <code>PERIODIC_ACT</code>
	 */
	final public static int PERIODIC_ACT = 1;

	/**
	 * The command is either specified as time interval or has a delay. This
	 * generates a command allowing to postpone its own execution either
	 * hardcoded or subjective to some conditions triggered by other BioDevices
	 * <code>DELAYED_ACT</code>
	 */
	final public static int DELAYED_ACT = 2;

	/**
	 * The command is a assembly of smaller Commands so as the resulting
	 * activity is a sequence of actions coordinated according to the individual
	 * rules of each single commands or according to global criteria defined in
	 * the command it self <code>DELAYED_ACT</code>
	 */
	final public static int MULTIPLE_ACT = 3;

	/**
	 * The size/value of single action is random. This applies only in the case
	 * where the command is asked to determine values to be sent to non targeted
	 * BioDevices which however are nto addressed by the parameters of the command
	 *  <code>RANDOM_MODE</code>
	 */
	final public static int RANDOM_VALUE_MODE = 0;

	/**
	 * The steps of the incrementing/changing are a finite inerval. This can be either
	 * expressed as a random value around a central point normalized via a modulo counter
	 * or a simpe truncation limited by upper and lower bounds of the intervall
	 * <code>RANGE_MODE</code>
	 */
	final public static int RRANGE_VALUE_MODE = 1;

	/**
	 * The execution of the command is either a cyclic random pattern or
	 * commands addressed from higher symbolic level. Such as the act of
	 * speaking or walking implicates that multiple commands and execution
	 * string must be sent repeatedly in cycle <code>REPEAT_MODE</code>
	 */
	final public static int REPEAT_VALUE_MODE = 2;

	/**
	 * @return Returns the command.
	 */
	public abstract Object[] getCommand();

	/**
	 * @param command
	 *            The command to set.
	 */
	public abstract void setCommand(Object[] command);

	/**
	 * external tool to modify the execution period
	 * 
	 * @param execPeriod
	 *            integer
	 */
	public abstract void setexecPeriod(int execPeriod);

	/**
	 * set a single instruction at index
	 * 
	 * @param o -
	 *            the instruction
	 * @param index -
	 *            the index of the instruction
	 */
	public abstract void setSingleCom(Object o, int index);

	/**
	 * @return - the array of individual commands
	 */
	public abstract Object[] getInstructions();

	/**
	 * @return - the period for this command
	 */
	public abstract int getPeriod();

	/**
	 * @return
	 */
	public abstract int getMode();

	/**
	 * @return - copy of this command
	 */
	public abstract Command cloneCommand();

	/**
	 * @return Returns the execMode.
	 */
	public abstract int getExecMode();

	/**
	 * @param execMode
	 *            The execMode to set.
	 */
	public abstract void setExecMode(int execMode);

	/**
	 * @return Returns the execType.
	 */
	public abstract int getExecType();

	/**
	 * @param execType
	 *            The execType to set.
	 */
	public abstract void setExecType(int execType);

	/**
	 * @return Returns the execDependency.
	 */
	public abstract int getExecDependency();

	/**
	 * @param execDependency
	 *            The execDependency to set.
	 */
	public abstract void setExecDependency(int execDependency);

	/**
	 * @return Returns the isNominal.
	 */
	public abstract boolean isNominal();

	/**
	 * @param isNominal
	 *            The isNominal to set.
	 */
	public abstract void setNominal(boolean isNominal);

	/**
	 * @return Returns the assignDevices.
	 */
	public abstract int getAssignDevices();

	/**
	 * @param assignDevices
	 *            The assignDevices to set.
	 */
	public abstract void setAssignDevices(int assignDevices);

	/**
	 * @return Returns the part.
	 */
	public abstract MNSRobotPart getPart();

	/**
	 * @param part
	 *            The part to set.
	 */
	public abstract void setPart(MNSRobotPart part);

	/**
	 * @return Returns the done.
	 */
	public abstract boolean isDone();

	/**
	 * @param done
	 *            The done to set.
	 */
	public abstract void setDone(boolean done);

	/**
	 * @return Returns the name.
	 */
	public abstract String getName();

	/**
	 * @param name The name to set.
	 */
	public abstract void setName(String name);
}