/*
 * Created on Feb 7, 2004
 */
package mns;

import biocell.BioThread;
import monitors.CommCenter;

/**
 * @author Martin Lukac
 */
abstract class CopyOfCommand extends BioThread implements Cloneable, BioDeviceListener {
	/**
	 * This execution is the simplest. We write the command and wait for
	 * feedback <code>SINGLE_ACT</code>
	 */
	final public static int SINGLE_ACT = 0;

	/**
	 * A periodic activity such as Speaking or Walking that requires a temporal
	 * synchronization of mulltiple BioDevices. This is done via the usage of
	 * the BioDeviceListener interface allowing to listen to each BioDevice
	 * Individualy <code>PERIODIC_ACT</code>
	 */
	final public static int PERIODIC_ACT = 1;

	/**
	 * The command is either specified as time interval or has a delay. This
	 * generates a command allowing to postpone its own execution either
	 * hardcoded or subjective to some conditions triggered by other BioDevices
	 * <code>DELAYED_ACT</code>
	 */
	final public static int DELAYED_ACT = 2;

	/**
	 * The command is a assembly of smaller Commands so as the resulting
	 * activity is a sequence of actions coordinated according to the individual
	 * rules of each single commands or according to global criteria defined in
	 * the command it self <code>DELAYED_ACT</code>
	 */
	final public static int MULTIPLE_ACT = 3;

	/**
	 * The size/value of single action is random. This applies only in the case
	 * where the command is asked to determine values to be sent to non targeted
	 * BioDevices which however are nto addressed by the parameters of the command
	 *  <code>RANDOM_MODE</code>
	 */
	final public static int RANDOM_VALUE_MODE = 0;

	/**
	 * The steps of the incrementing/changing are a finite inerval. This can be either
	 * expressed as a random value around a central point normalized via a modulo counter
	 * or a simpe truncation limited by upper and lower bounds of the intervall
	 * <code>RANGE_MODE</code>
	 */
	final public static int RRANGE_VALUE_MODE = 1;

	/**
	 * The execution of the command is either a cyclic random pattern or
	 * commands addressed from higher symbolic level. Such as the act of
	 * speaking or walking implicates that multiple commands and execution
	 * string must be sent repeatedly in cycle <code>REPEAT_MODE</code>
	 */
	final public static int REPEAT_VALUE_MODE = 2;

	public CopyOfCommand(String name) {
		super(name);
		this.name = name;
	}

	/**
	 * @return Returns the command.
	 */
	public Object[] getCommand() {
		return command;
	}

	/**
	 * @param command
	 *            The command to set.
	 */
	public void setCommand(Object[] command) {
		this.command = command;
	}

	/**
	 * external tool to modify the execution period
	 * 
	 * @param execPeriod
	 *            integer
	 */
	public void setexecPeriod(int execPeriod) {
		this.execPeriod = execPeriod;
		this.execType = REPEAT_VALUE_MODE;
	}

	/**
	 * set a single instruction at index
	 * 
	 * @param o -
	 *            the instruction
	 * @param index -
	 *            the index of the instruction
	 */
	public void setSingleCom(Object o, int index) {
		if (index < this.command.length)
				this.command[index] = o;
	}


	/**
	 * @return - the array of individual commands
	 */
	public Object[] getInstructions() {
		return this.command;
	}

	/**
	 * @return - the period for this command
	 */
	public int getPeriod() {
		return this.execPeriod;
	}

	/**
	 * @return
	 */
	public int getMode() {
		return this.execMode;
	}

	/**
	 * @return - copy of this command
	 */
	public CopyOfCommand cloneCommand() {
		try {
			return (CopyOfCommand) this.clone();
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * @return Returns the execMode.
	 */
	public int getExecMode() {
		return execMode;
	}

	/**
	 * @param execMode
	 *            The execMode to set.
	 */
	public void setExecMode(int execMode) {
		this.execMode = execMode;
	}

	/**
	 * @return Returns the execType.
	 */
	public int getExecType() {
		return execType;
	}

	/**
	 * @param execType
	 *            The execType to set.
	 */
	public void setExecType(int execType) {
		this.execType = execType;
	}

	/**
	 * @return Returns the execDependency.
	 */
	public int getExecDependency() {
		return execDependency;
	}

	/**
	 * @param execDependency
	 *            The execDependency to set.
	 */
	public void setExecDependency(int execDependency) {
		this.execDependency = execDependency;
	}

	/**
	 * @return Returns the isNominal.
	 */
	public boolean isNominal() {
		return isNominal;
	}

	/**
	 * @param isNominal
	 *            The isNominal to set.
	 */
	public void setNominal(boolean isNominal) {
		this.isNominal = isNominal;
	}

	/**
	 * @return Returns the assignDevices.
	 */
	public int getAssignDevices() {
		return assignDevices;
	}

	/**
	 * @param assignDevices
	 *            The assignDevices to set.
	 */
	public void setAssignDevices(int assignDevices) {
		this.assignDevices = assignDevices;
	}

	/**
	 * @return Returns the part.
	 */
	public MNSRobotPart getPart() {
		return part;
	}

	/**
	 * @param part
	 *            The part to set.
	 */
	public void setPart(MNSRobotPart part) {
		this.part = part;
	}

	/**
	 * @return Returns the done.
	 */
	public boolean isDone() {
		return done;
	}

	/**
	 * @param done
	 *            The done to set.
	 */
	synchronized public void setDone(boolean done) {
		this.done = done;
		notifyAll();
	}

	/**
	 * private method for adding this to the biodevices as a listener
	 *  
	 */
	protected void addListener() {
		for (int a = 0; a < getPart().getDevices().length; a++)
			if (getPart().getDevice(a) != null) {
				getPart().getDevice(a).addListener(this);
				CommCenter.showDebug("Listener added");
			}
	}

	/*******************************************************
	 * This section defines runtime variables for handling the commands
	 * during operations and processing 
	 *******************************************************/
	/**
	 * The name of the command <code>name</code>
	 */
	private String name = null;

	/**
	 * the command vector representing channels to BioDevices so as the
	 * instruction set . According to the type of of implemented command
	 * this object holds either:
	 *	only String instructions representing individual command 
	 *	instructions for the BioDevices
	 *either:
	 *	a set of other commands to be executed according to the defined 
	 *	mode of operation of this command<code>command</code>
	 */
	private Object[] command;

	/**
	 * <code>execMode</code> Mode of the command indicates the way of
	 * generating certain type of commands. in the RANDOM_VALUE_MODE any values
	 * non defined in a command will be randomly replaced , in the
	 * RRANGE_VALUE_MODE the generation of auxilliary variables will be still
	 * random but only in a determined interval, finally REPEAT_VALUE_MODE
	 * allows to repeat any values of the command instruction chain
	 *  
	 */
	private int execMode = RANDOM_VALUE_MODE;

	/**
	 * Temporal type of execution can be of three primitive values: SINGLE_ACT,
	 * PERIODIC_ACT, DELAYED_ACT <code>execType</code>
	 */
	private int execType = SINGLE_ACT;

	/**
	 * Time period for repeating behavior of the command. This is only a indicator 
	 * and in the future will be useless because periodc command generation can be
	 * replaced by the synchronization of multiple BioDevices giving rise to a 
	 * multi-coordinated activity.
	 *  <code>execPeriod</code>
	 */
	private int execPeriod = 0;

	/**
	 * <code>the dependency of execution</code>- it is possible to move more
	 * devices and keep the ratios between them
	 */
	private int execDependency = 0;

	/**
	 * Comment for
	 * <code>priority of the command 0 - the lowest 3 - the highest</code>
	 */
	private int cPriority = 0;

	/**
	 * This indicates how to treat the command. A nominal command
	 * is executed exactly as it was stored when created. All arguments are
	 * generally ingnored. A non-nominal command will be executed as a normal command;
	 * i.e. the given parameters will be replaced in the original command and 
	 * invalid or missing parameters will be generated according to the execMode
	 * <code>isNominal</code>
	 */
	private boolean isNominal = false;

	/**
	 * Counter used durring the execution and synchronization of activity
	 * within one single command. The processing of every single command
	 * will be checking the number of devices in order to define when the 
	 * command is done<code>assignDevices</code>
	 */
	private int assignDevices = 0;

	/**
	 * Indicator of the activity of this Command. This is strictly a internal indicator
	 * and shall not be used outside of the internal code of a individual Command
	 * object. <code>done</code>
	 */
	private boolean done = false;

	/**
	 * The listener and the RobotPart that generated this command is directly imported
	 * the Command object in order to allow certian references and communication. 
	 * This allows not to import listeners of every single concerned BioDevice and
	 * makes the Code for the Command smaller in terms of allocated memory.
	 * <code>part</code>
	 */
	private MNSRobotPart part = null;

}