/************************************************************************
 *	Martin Lukac, 2003-2005 		*
 *	Intelligent Robotics Laboratory										*
 *	Portland State University 											*
 *	Office: 70-08 FAB, phone: 503-490-8478								*
 *	Mail:	P.O. Box 751												*
 *			Portland, OR 97207-0751										*
 *	Email:	lukacm@cecs.pdx.edu										*
 *	Updates and documentation: 	www.pdx.edu\~lukacm				*
 ************************************************************************/
// ************************************************************************
// * 	This program is provided WITHOUT ANY WARRANTY either expressed
// *	or implied.  You may study, use, modify, and distribute it for
// *	any non-commercial purposes.  For any commercial use, prior 
// *	permission from the authors must be granted.
// ************************************************************************/
/*
 * Created on Feb 13, 2004
 */
package mns;

import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.InputStreamReader;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;

import mainframe.Tools;
import monitors.CommCenter;
import biocell.BioThread;
import biocell.DynamicMood;
import devs.motorpackage.asc16.Servo;

/**
 * 
 * Class implementing parts of the robot from the symbolic or highest point of
 * view Each BodyPart has a set of associated devices that are fetched against
 * the incoming commands in form of either files or strings This set of command
 * is send to CNS that takes care about the synchronization and other low level
 * features
 * 
 * The order of devices is as follows: first is speech synthesis second is
 * speech recognition Third is vision control system all devices after are
 * servos associated with robot movements
 * 
 * Each CNSRobotPart can have only one ThreadedCommand at a time This is not
 * restricted but the feedback is not tracking the appropriatte devices and
 * confusion between threads can occur This is done by initiating a command with
 * a ExecPeriod != 0 ThreadedCommand will be executing all parts of the command
 * beside SpeechProduction.
 * 
 * @author Martin Lukac
 * 
 */
public class MNSRobotPart extends BioThread implements CommandListener {

    /*
     * (non-Javadoc)
     * 
     * @see mns.CommandListener#commandOccured(mns.CommandEvent)
     */
    public void commandOccured(CommandEvent o) {
        // if (o.getMessage().toLowerCase().startsWith("done"))
        listener.stateChanged(new MNSRobotPartEvent(BioDevice.DONE, this));

    }

    /**
     * @param RobotPartListener -
     *            the CNSRobotPart listener
     * @param name -
     *            the name of the command
     */
    public MNSRobotPart(MNSRobotPartListener listener, String name) {
        super(name);
        this.name = name;
        this.listener = listener;
        /** setting the current (default) idle state */
        currentState = BioDevice.DONE;
        /** default number of commands for this RobotPart */
        start();
    }

    /**
     * @param RobotPartListener -
     *            the CNSRobotPart listener
     * @param devs -
     *            the pointers to dedicated devices
     * @param name -
     *            the name of the robotpart
     */
    public MNSRobotPart(MNSRobotPartListener listener, BioDevice[] devs,
            String name) {
        super(name);
        this.name = name;
        this.listener = listener;
        currentState = BioDevice.DONE;
        /** adds the devices to this robot part */
        addDevs(devs);
        start();
    }

    /**
     * Initiate the RobotPArt with a initial set of control commands. These re
     * implemented inside of each biodevice in the loadCNTR(); These commands
     * can be extended very easily by enlarging the size of the list file
     */
    private void init() {
        try {
            File f = new File("mns/mns/default.comm");
            String line;
            BufferedReader bfr = new BufferedReader(new FileReader(f));
            // set the initial set of control commands
            while ((line = bfr.readLine()) != null) {
                if (line.length() < 2)
                    break;
                StringTokenizer t = new StringTokenizer(line, ",");
                String a[] = new String[2];
                a[0] = t.nextToken();
                String key = t.nextToken();
                String data = "";
                for (int g = 0; g < devs.length; g++) {
                    data += "null ";
                }
                a[1] = data;
                CommandThread command = new CommandThread(a, key.charAt(0));
                commands.add(command);
                command.setAssignDevices(devs.length);
                command.setPart(this);
                CommCenter.showDebug(this.name + ": added primitive: " + a[0]+ " " +a[1]);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * add a command to the list of current commands, the ordering of the
     * variables for the constructor matters!!!! Thus synax is maximally
     * required.
     * 
     * @param command -
     *            command to be added
     */
    public synchronized void addCommand(String biocommand) {
        while (buildingBusy) {
            try {
                sleep(50);
            } catch (InterruptedException e) {

            }
        }
        setBuildingBusy(true);
        String[] params = Tools.tokenize(biocommand, ":");
        // the speech command is particulalry distinguished
        // this does not require any specifications
        String temp;
        try {
            temp = params[0].toLowerCase().substring(0, 4);
        } catch (Exception e) {
            temp = params[0].toLowerCase().substring(0);
        }
        if (isCommand(params[0].toLowerCase().trim()) != true) {
            if (temp.startsWith("spea") || temp.startsWith("talk")
                    || temp.startsWith("spee")) {
                Command command = new SpeachThread(params[0]);
                CommCenter.showDebug("Creating speech command: " + params[0]);
                commands.add(command);
                command.setAssignDevices(devs.length);
                command.setPart(this);
            } else if ((biocommand.indexOf(':') < biocommand.lastIndexOf(":"))
                    || params.length - 1 > getDevices().length) {
                // the complex commands allows a sequence of commands to be
                // executed
                CommCenter.showDebug("Creating complex command: " + biocommand);
                StringTokenizer t = new StringTokenizer(biocommand, ":");
                String me = t.nextToken();
                Vector subs = new Vector(1, 1);
                while (t.hasMoreTokens()) {
                    String to = ((String)t.nextToken()).trim();
                    String na;
                    boolean simple = false;
                    boolean single = false;
                    
                    //check wether there is only one argument
                    try {
                        na = to.substring(0, to.indexOf(' '));
                        single = false;
                    }catch (Exception e){
                        na = to;
                        single = true;
                    }
                    //check if the command is a simplified move command
                    try{
                        if (na.startsWith("null") || na.startsWith("ignor")){
                            System.out.println("Parsed simple: "+na);
                            na = "move";
                            simple = true;
                        }else {
                            int a = Integer.parseInt(na);
                            System.out.println("Parsed simple: "+a);
                            na = "move";
                            simple = true;
                        }
                    } catch (Exception e){
                        simple = false;
                        //check if this is not a nominal command - single word
                        if (single)
                            to = "";
                        else to = to.substring(to.indexOf(' ')+1);
                    }
                    
                    
                    
                    
                    String[] pars = { na, to };
                    Command fg = getCommand(na);
                    CommandThread comm;
                    if (fg == null){
                        System.out.println("Creating Subroutine: "+ na +"  "+to);
                        comm = new CommandThread(pars);
                    } else {
                        System.out.println("Cloning Subroutine: "+ fg.getName() +"  "+to);
                        comm = (CommandThread)fg.cloneCommand();
                        String [] mnt = {to};
                        comm.setCommand(mnt);
                    }
                    comm.setAssignDevices(devs.length);
                    subs.add(comm);
                }
                Command command = new ComplexCommand(me, subs);
                commands.add(command);
                command.setPart(this);

            } else {
                // finally simple command is just a one use sequence defined in
                // time
                String com = params[0].toLowerCase().trim();
                CommandThread command = null;
                command = new CommandThread(params);
                commands.add(command);
                command.setAssignDevices(devs.length);
                command.setPart(this);
                CommCenter.showDebug("Creating simple threaded command: "
                        + biocommand + " on " + devs.length + " devices");
            }
        }
        setBuildingBusy(false);
        notifyAll();
    }

    synchronized private boolean isCommand(String command) {
        System.out.println("Comparing "+command);
        Iterator i = commands.iterator();
        while (i.hasNext()) {
            String s = (((Command) i.next()).getName()).trim();
            System.out.print("'"+s+"', ");
            if (s.equalsIgnoreCase(command))
                return true;
        }
        System.out.println();
        return false;
    }

    /**
     * get a Command if available in this CNSRobotPart
     * 
     * @param commandName -
     *            name of the desired command
     * @return - returns null or the command found
     */
    public Command getCommand(String comName) {
        Iterator i = commands.iterator();
        stepEnergy(comName);
        while (i.hasNext()) {
            try {
                Command comm = ((Command) i.next());
                if (comm.getBioname().trim().equalsIgnoreCase(comName.trim()))
                    return comm;
            } catch (Exception e) {
                StackTraceElement ele[] = e.getStackTrace();
                String s = "";
                for (int a = 0; a < ele.length; a++)
                    s += ele[a].toString() + "\n";
                CommCenter.showError("MNSRobotPart " + s);
                return null;
            }
        }
        return null;
    }

    public Vector getCommands() {
        stepEnergy(commands);
        return commands;
    }

    /** @return - the list of biodevices associated with this robotpart */
    public BioDevice[] getDevices() {
        stepEnergy(devs);
        return devs;
    }

    /**
     * return the BioDevice at specified index
     * 
     * @param index -
     *            index of the device
     * @return - BioDevice
     */
    public BioDevice getDevice(int index) {
        try {
            stepEnergy(devs[index]);

            return this.devs[index];
        } catch (NullPointerException e) {
            return null;
        }
    }

    /**
     * add a set of devices (BioDevices) and replace the old ones
     * 
     * @param devs -
     *            the array of BioDevices
     */
    public void addDevs(BioDevice[] devs) {
        stepEnergy(devs);

        this.devs = devs;
        init();

    }

    /**
     * @param bioDevices
     */
    public void addDevs(Vector bioDevices) {
        stepEnergy(bioDevices);
        devs = (BioDevice[]) bioDevices.toArray();
        init();

    }

    /**
     * get the name of this CNSRobotPArt
     * 
     * @return - String - name of the comp0nent
     */
    public String getname() {
        stepEnergy(name);
        return name;
    }

    /**
     * set name of this CNSRobotPart
     * 
     * @param name -
     *            String
     */
    public void setname(String name) {
        stepEnergy(name);
        this.name = name;
    }

    public void setPartMax(int max, int bioindex) {
        System.out.println(this.getname() + " sets the Servo Max: " + max);
        if (bioindex < devs.length)
            ((Servo) devs[bioindex]).setNormHighPosition(max);
    }

    public void setPartMin(int min, int bioindex) {
        System.out.println(this.getname() + " sets the Servo Min: " + min);
        if (bioindex < devs.length)
            ((Servo) devs[bioindex]).setNormLowPosition(min);
    }

    public void setPartSpeed(int speed, int bioindex) {
        CommCenter.showDebug(speed + " index: " + bioindex);
        if (bioindex < devs.length)
            ((Servo) devs[bioindex]).setSpeed(speed);
    }

    public int getPartAccel(int bioindex) {
        if (bioindex < devs.length)
            return ((Servo) devs[bioindex]).getAcceleration();// getAcceleration();
        else
            return 0;
    }

    public int getPartMax(int bioindex) {
        if (bioindex < devs.length)
            return ((Servo) devs[bioindex]).getNormHighPosition();
        else
            return 0;
    }

    public int getPartMin(int bioindex) {
        if (bioindex < devs.length)
            return ((Servo) devs[bioindex]).getNormLowPosition();
        else
            return 0;
    }

    public int getPartSpeed(int bioindex) {
        if (bioindex < devs.length)
            return ((Servo) devs[bioindex]).getCurrentSpeed();
        else
            return 0;
    }

    public void setPartAccel(int accel, int bioindex) {
        CommCenter.showDebug(accel + " index: " + bioindex);
        if (bioindex < devs.length)
            ((Servo) devs[bioindex]).setAcceleration(accel);
    }

    /**
     * @return Returns the buildingBusy.
     */
    private synchronized boolean isBuildingBusy() {
        return buildingBusy;
    }

    /**
     * @param buildingBusy
     *            The buildingBusy to set.
     */
    private synchronized void setBuildingBusy(boolean buildingBusy) {
        this.buildingBusy = buildingBusy;
    }

    /**
     * @param max
     */
    public void setPartMax(String max) {
        stepEnergy(max);
        try {
            CommCenter.showDebug(" Seting RobotPart Maximum: " + max);
            String[] indx = Tools.tokenize(max);
            if (name.equalsIgnoreCase("mouth")) {
                for (int a = devs.length - 1; a > 0; a--) {
                    ((Servo) devs[a]).setNormHighPosition(Integer
                            .parseInt(indx[a]));
                    CommCenter.showDebug(getname() + " Maximum set at: "
                            + ((Servo) devs[a]).getNormHighPosition());
                }
            } else
                for (int a = 0; a < devs.length; a++) {
                    ((Servo) devs[a]).setNormHighPosition(Integer
                            .parseInt(indx[a]));
                    CommCenter.showDebug(getname() + " Maximum set at: "
                            + ((Servo) devs[a]).getNormHighPosition());
                }
        } catch (Exception e) {
            StackTraceElement ele[] = e.getStackTrace();
            String s = "";
            for (int a = 0; a < ele.length; a++)
                s += ele[a].toString() + "\n";
            CommCenter.showError("MNSRobotPart " + s);
        }
    }

    /**
     * @param min
     */
    public void setPartMin(String min) {
        stepEnergy(min);
        try {
            CommCenter.showDebug(" Seting RobotPart Minimum: " + min);
            String[] indx = Tools.tokenize(min);
            if (name.equalsIgnoreCase("mouth")) {
                for (int a = devs.length - 1; a > 0; a--) {
                    ((Servo) devs[a]).setNormLowPosition(Integer
                            .parseInt(indx[a]));
                    CommCenter.showDebug(getname() + " Minimum set at: "
                            + ((Servo) devs[a]).getNormLowPosition());
                }
            } else
                for (int a = 0; a < devs.length; a++) {
                    ((Servo) devs[a]).setNormLowPosition(Integer
                            .parseInt(indx[a]));
                    CommCenter.showDebug(getname() + " Minimum set at: "
                            + ((Servo) devs[a]).getNormLowPosition());
                }
        } catch (Exception e) {
            StackTraceElement ele[] = e.getStackTrace();
            String s = "";
            for (int a = 0; a < ele.length; a++)
                s += ele[a].toString() + "\n";
            CommCenter.showError("MNSRobotPart " + s);
        }
    }

    /**
     * fetch the current state according to BioDevice scheme - unusual but
     * exploitable in the future
     * 
     * @return - the current state of the RobotPart
     */
    public int getstate() {
        stepEnergy(Integer.toString(currentState));
        return currentState;
    }

    /**
     * The only public method to execute commands from a external parent such as
     * the CNS All commands must be previously registered specific commands for
     * control of devices: parameters modifying the way of doing things in short
     * direct commands to the hardware that needs a thread
     * 
     * executes a command if it exists with the set of parameters
     * 
     * @param comm -
     *            name of the command
     * @param argument -
     *            argument to the command
     */
    synchronized public void execute(String comm, String argument) {
        Command c = getCommand(comm);
        if (c != null)
            stepEnergy(c);
        else {
            CommCenter.showDebug(this.name + " error - no such a command: "
                    + comm);
            notifyAll();
            return;
        }

        Command nc = c.cloneCommand();
        if (nc.isNominal()) {
            commandQueue.add(nc);
            notifyAll();
        } else {
            /**
             * for all other commands just parse the argument and set it as
             * argument to the command
             */
            try {
                if (nc.getBioname().toLowerCase().startsWith("speech")) {
                    CommCenter.showDebug(" Speech: " + argument);
                    nc.setSingleCom(argument, 0);
                } else {
                    System.out.println(c + " : " + comm + " < new > "
                            + argument);
                    if (argument != null && argument.length() > 0) {
                        nc.setSingleCom(argument, 0);
                    }
                }
                CommCenter.showDebug(this.name + ": started " + nc.getName()
                        + " " + argument);
                // add the command to th queue
                commandQueue.add(nc);
            } catch (NullPointerException e) {
                e.printStackTrace();
            }
            notifyAll();
        }
    }

    /**
     * overriden from Thread Each RobotPart is sleeping for a period = endCycle
     * and then checks for the command Queue. If required the command will start
     * a new Thread generating impulses for some cyclic movements
     */
    public void run() {
        long time = 0;
        try {
            synchronized (this) {
                while (true) {
                    /** waits until the commandQueue is empty */
                    while (commandQueue == null || commandQueue.size() <= 0)
                        try {
                            wait((long) (1000 * getMoodParam("waittime")));
                        } catch (InterruptedException e) {
                            System.out.println("awaken");
                        }

                    // gets and executes the first command
                    CommCenter.showDebug(this.getname() + " is executing "
                            + ((Command) commandQueue.get(0)).getBioname());
                    ((Command) commandQueue.remove(0)).start();
                }
            }
        } catch (Exception e) {
            StackTraceElement ele[] = e.getStackTrace();
            String s = "";
            for (int a = 0; a < ele.length; a++)
                s += ele[a].toString() + "\n";
            CommCenter.showError("MNSRobotPart " + s);
        }
    }

    /**
     * @param currentState
     */
    synchronized public void stateChanged(int currentState) {
        this.currentState = currentState;
        if (listener != null && currentState == BioDevice.DONE) {
            listener.stateChanged(new MNSRobotPartEvent(currentState, this));
        }
    }

    /** <code>name of this CNSRobotPart </code> */
    private String name = null;

    /** <code>the vector of commands affecting this body part </code> */
    private Vector commands = new Vector(1, 1);

    /** <code>the control CNS - up stream data flow </code> */
    private MNSRobotPartListener listener;

    /** <code>the pointers to functional bioDevices </code> */
    private BioDevice[] devs;

    /** <code>currentState</code> */
    private int currentState;

    /** <code>command queue </code> */
    private List commandQueue = new Vector(1, 1);

    private boolean buildingBusy = false;
}
